/*
 * src/c_dd.cc
 *
 * This work was supported by the Director, Office of Science, Division
 * of Mathematical, Information, and Computational Sciences of the
 * U.S. Department of Energy under contract number DE-AC03-76SF00098.
 *
 * Copyright (c) 2000-2001
 *
 * Contains the C wrapper functions for double-double precision arithmetic.
 * This can be used from Fortran code.
 */

/* Translated to pure C code by T.Kouya */
/* 2015-03-01 */

#ifndef __C_DD_QD_H
	#define __C_DD_QD_H

#define _QD_SPLITTER 134217729.0               // = 2^27 + 1
#define _QD_SPLIT_THRESH 6.69692879491417e+299 // = 2^996

#ifdef QD_VACPP_BUILTINS_H
/* For VisualAge C++ __fmadd */
#include <builtins.h>
#endif

#include <math.h>
#include <limits.h>

#define TO_DOUBLE_PTR(a, ptr) ptr[0] = a.x[0]; ptr[1] = a.x[1];

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Set the round-to-double flag, and save the old control word in old_cw.
 * If old_cw is NULL, the old control word is not saved.
 */
void fpu_fix_start(unsigned int *old_cw);

/*
 * Restore the control word.
 */
void fpu_fix_end(unsigned int *old_cw);

/**************************************/
/* DD                                 */
/**************************************/
//  double _hi() const { return x[0]; }
//  double _lo() const { return x[1]; }
#define DD_HI(a)	((a)[0])
#define DD_LOW(a)	((a)[1])

// Definitions for Comparing Functions
typedef unsigned int	dd_bool;
#define DD_TRUE		(1UL)
#define DD_FALSE	(0UL)

#define DD_ISNAN(a) (isnan((a)[0]))
#define QD_ISNAN(a) (isnan((a)[0]))

#define DD_ISINF(a) (isinf((a)[0]))
#define QD_ISINF(a) (isinf((a)[0]))

#define DD_ISZERO(a) ((a)[0] == 0.0)
#define QD_ISZERO(a) ((a)[0] == 0.0)

#define DD_ISNEGATIVE(a) ((a)[0] < 0.0)
#define QD_ISNEGATIVE(a) ((a)[0] < 0.0)

#define DD_NAN (FP_NAN)
#define QD_NAN (FP_NAN)

// PI = 3.1415...
static const double const_qd_pi[4] = {3.067961575771282340e-03, 1.195944139792337116e-19, -2.924579892303066080e-36, 1.086381075061880158e-52};
static const double const_dd_pi[2] = {1.963495408493620697e-01, 7.654042494670957545e-18};

double to_double(double a);
int    to_int(double a);

/*********** Basic Functions ************/
/* Computes fl(a+b) and err(a+b).  Assumes |a| >= |b|. */
double quick_two_sum(double a, double b, double *err);

/* Computes fl(a-b) and err(a-b).  Assumes |a| >= |b| */
double quick_two_diff(double a, double b, double *err);

/* Computes fl(a+b) and err(a+b).  */
double two_sum(double a, double b, double *err);

/* Computes fl(a-b) and err(a-b).  */
double two_diff(double a, double b, double *err);

#ifndef QD_FMS
/* Computes high word and lo word of a */
void split(double a, double *hi, double *lo);
#endif

/* Computes fl(a*b) and err(a*b). */
double two_prod(double a, double b, double *err);

/* Computes fl(a*a) and err(a*a).  Faster than the above method. */
double two_sqr(double a, double *err);

/* Computes the nearest integer to d. */
double nint(double d);

/* Computes the truncated integer. */
double aint(double d);

/* These are provided to give consistent 
   interface for double with double-double and quad-double. */
void sincosh(double t, double *sinh_t, double *cosh_t);

double sqr(double t);

//#ifdef NATIVE_C
/* double-double = double + double */
void c_d_add(double a, double b, double *c);

/* add */
void c_dd_add(const double *a, const double *b, double *c);

// sloppy_add
void c_dd_add_sloppy(const double *a, const double *b, double *c, double *err);

void c_dd_add_dd_d(const double *a, double b, double *c);

void c_dd_add_d_dd(double a, const double *b, double *c);

/*********** Subtractions ************/
/* double-double = double - double */
void c_d_sub(double a, double b, double *c);

/* sub */
void c_dd_sub(const double *a, const double *b, double *c);

void c_dd_sub_sloppy(const double *a, const double *b, double *c);

/* double-double - double */
void c_dd_sub_dd_d(const double *a, double b, double *c);

/* double - double-double */
void c_dd_sub_d_dd(double a, const double *b, double *c);

/*********** Multiplications ************/
/* double-double = double * double */
void  c_d_mul(double a, double b, double *c);

/* mul */
void c_dd_mul(const double *a, const double *b, double *c);

void c_dd_mul_dd_d(const double *a, double b, double *c);

void c_dd_mul_d_dd(double a, const double *b, double *c);

/*********** Divisions ************/
void c_d_div(double a, double b, double *c);

/* div */
void c_dd_div(const double *a, const double *b, double *c);

/* double-double / double-double */
void c_dd_sloppy_div(double *a, double *b, double *c);

/* double-double / double */
void c_dd_div_dd_d(const double *a, double b, double *c);

/* double / double-double */
void c_dd_div_d_dd(double a, const double *b, double *c);

/* copy */
void c_dd_copy(const double *a, double *b);

void c_dd_copy_d(double a, double *b);

/* b := a^2 */
void c_d_sqr(double a, double *b);

/* b := a^2 */
void c_dd_sqrt(const double *a, double *b);

// b := a^2
void c_dd_sqr_d(double a, double *ret);

// b := a^2
void c_dd_sqr(const double *a, double *b);

void c_dd_abs(const double *a, double *b);

void c_dd_npwr(const double *a, int n, double *b);

void c_dd_nroot(const double *a, int n, double *b);

void c_dd_nint(const double *a, double *b);

void c_dd_aint(const double *a, double *b);

void c_dd_floor(const double *a, double *b);

void c_dd_ceil(const double *a, double *b);

void c_dd_log(const double *a, double *b);

void c_dd_log10(const double *a, double *b);

void c_dd_exp(const double *a, double *b);

void c_dd_sin(const double *a, double *b);

void c_dd_cos(const double *a, double *b);

void c_dd_tan(const double *a, double *b);

void c_dd_asin(const double *a, double *b);

void c_dd_acos(const double *a, double *b);

void c_dd_atan(const double *a, double *b);

void c_dd_atan2(const double *a, const double *b, double *c);

void c_dd_sinh(const double *a, double *b);

void c_dd_cosh(const double *a, double *b);

void c_dd_tanh(const double *a, double *b);

void c_dd_asinh(const double *a, double *b);

void c_dd_acosh(const double *a, double *b);

void c_dd_atanh(const double *a, double *b);

void c_dd_sincos(const double *a, double *s, double *c);

void c_dd_sincosh(const double *a, double *s, double *c);

void c_dd_read(const char *s, double *a);

void c_dd_swrite(const double *a, int precision, char *s, int len);

void c_dd_write(const double *a);

void c_dd_neg(const double *a, double *b);

void c_dd_rand(double *a);

void c_dd_comp(const double *a, const double *b, int *result);

void c_dd_comp_dd_d(const double *a, double b, int *result);

void c_dd_comp_d_dd(double a, const double *b, int *result);

void c_dd_pi(double *a);

/*********** Micellaneous ************/
/*  this == 0 */
//bool is_zero(double *x);
dd_bool dd_is_zero(const double *x);

/*  this == 1 */
//bool is_one(double *x);
dd_bool dd_is_one(const double *x);

/*  this > 0 */
//bool is_positive(double *x);
dd_bool dd_is_positive(const double *x);

/* this < 0 */
//bool is_negative(double *x);
dd_bool dd_is_negative(const double *x);

/**************************************/
/* QD                                 */
/**************************************/

/* add */
void c_qd_add(const double *a, const double *b, double *c);
void c_qd_add_dd_qd(const double *a, const double *b, double *c);
void c_qd_add_qd_dd(const double *a, const double *b, double *c);
void c_qd_add_d_qd(double a, const double *b, double *c);
void c_qd_add_qd_d(const double *a, double b, double *c);
void c_qd_selfadd(const double *a, double *b);
void c_qd_selfadd_dd(const double *a, double *b);
void c_qd_selfadd_d(double a, double *b);

/* sub */
void c_qd_sub(const double *a, const double *b, double *c);
void c_qd_sub_dd_qd(const double *a, const double *b, double *c);
void c_qd_sub_qd_dd(const double *a, const double *b, double *c);
void c_qd_sub_d_qd(double a, const double *b, double *c);
void c_qd_sub_qd_d(const double *a, double b, double *c);
void c_qd_selfsub(const double *a, double *b);
void c_qd_selfsub_dd(const double *a, double *b);
void c_qd_selfsub_d(double a, double *b);

/* mul */
void c_qd_mul(const double *a, const double *b, double *c);
void c_qd_mul_dd_qd(const double *a, const double *b, double *c);
void c_qd_mul_qd_dd(const double *a, const double *b, double *c);
void c_qd_mul_d_qd(double a, const double *b, double *c);
void c_qd_mul_qd_d(const double *a, double b, double *c);
void c_qd_selfmul(const double *a, double *b);
void c_qd_selfmul_dd(const double *a, double *b);
void c_qd_selfmul_d(double a, double *b);

/* div */
void c_qd_div(const double *a, const double *b, double *c);
void c_qd_div_dd_qd(const double *a, const double *b, double *c);
void c_qd_div_qd_dd(const double *a, const double *b, double *c);
void c_qd_div_d_qd(double a, const double *b, double *c);
void c_qd_div_qd_d(const double *a, double b, double *c);
void c_qd_selfdiv(const double *a, double *b);
void c_qd_selfdiv_dd(const double *a, double *b);
void c_qd_selfdiv_d(double a, double *b);

/* copy */
void c_qd_copy(const double *a, double *b);
void c_qd_copy_dd(const double *a, double *b);
void c_qd_copy_d(double a, double *b);

void c_qd_sqrt(const double *a, double *b);
void c_qd_sqr(const double *a, double *b);

void c_qd_abs(const double *a, double *b);

void c_qd_npwr(const double *a, int b, double *c);
void c_qd_nroot(const double *a, int b, double *c);

void c_qd_nint(const double *a, double *b);
void c_qd_aint(const double *a, double *b);
void c_qd_floor(const double *a, double *b);
void c_qd_ceil(const double *a, double *b);

void c_qd_exp(const double *a, double *b);
void c_qd_log(const double *a, double *b);
void c_qd_log10(const double *a, double *b);

void c_qd_sin(const double *a, double *b);
void c_qd_cos(const double *a, double *b);
void c_qd_tan(const double *a, double *b);

void c_qd_asin(const double *a, double *b);
void c_qd_acos(const double *a, double *b);
void c_qd_atan(const double *a, double *b);
void c_qd_atan2(const double *a, const double *b, double *c);

void c_qd_sinh(const double *a, double *b);
void c_qd_cosh(const double *a, double *b);
void c_qd_tanh(const double *a, double *b);

void c_qd_asinh(const double *a, double *b);
void c_qd_acosh(const double *a, double *b);
void c_qd_atanh(const double *a, double *b);

void c_qd_sincos(const double *a, double *s, double *c);
void c_qd_sincosh(const double *a, double *s, double *c);

void c_qd_read(const char *s, double *a);
void c_qd_swrite(const double *a, int precision, char *s, int len);
void c_qd_write(const double *a);
void c_qd_neg(const double *a, double *b);
void c_qd_rand(double *a);
void c_qd_comp(const double *a, const double *b, int *result);
void c_qd_comp_qd_d(const double *a, double b, int *result);
void c_qd_comp_d_qd(double a, const double *b, int *result);
void c_qd_pi(double *a);

#ifdef __cplusplus
}
#endif

#endif // __C_DD_QD_H
